/*
 * Copyright (C) 2003 Andras Varga; CTIE, Monash University, Australia
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "inet/applications/smartGrid/protocol/EtherAppGoose.h"

#include "inet/applications/ethernet/EtherApp_m.h"
#include "inet/common/ModuleAccess.h"
#include "inet/common/TimeTag_m.h"
#include "inet/common/packet/Packet.h"
#include "inet/linklayer/common/Ieee802Ctrl.h"
#include "inet/linklayer/common/Ieee802SapTag_m.h"
#include "inet/linklayer/common/MacAddressTag_m.h"
#include "inet/networklayer/common/L3AddressResolver.h"

namespace inet {

Define_Module(EtherAppGoose);

EtherAppGoose::~EtherAppGoose()
{
    cancelAndDelete(timerMsg);
}

void EtherAppGoose::initialize(int stage)
{
    if (stage == INITSTAGE_APPLICATION_LAYER && isGenerator())
        timerMsg = new cMessage("generateNextPacket");

    ApplicationBase::initialize(stage);

    if (stage == INITSTAGE_LOCAL) {
        reqLength = &par("reqLength");
        respLength = &par("respLength");
        sendInterval = &par("sendInterval");

        localSap = par("localSAP");
        remoteSap = par("remoteSAP");

        seqNum = 0;
        WATCH(seqNum);

        // statistics
        packetsSent = packetsReceived = 0;
        WATCH(packetsSent);
        WATCH(packetsReceived);

        startTime = par("startTime");
        stopTime = par("stopTime");
        if (stopTime >= SIMTIME_ZERO && stopTime < startTime)
            throw cRuntimeError("Invalid startTime/stopTime parameters");
        llcSocket.setOutputGate(gate("out"));
        llcSocket.setCallback(this);
    }
}

void EtherAppGoose::handleMessageWhenUp(cMessage *msg)
{
    if (msg->isSelfMessage()) {
        if (msg->getKind() == START) {
            EV_DEBUG << getFullPath() << " registering DSAP " << localSap << "\n";
            llcSocket.open(-1, localSap);

            destMacAddress = resolveDestMacAddress();
            // if no dest address given, nothing to do
            if (destMacAddress.isUnspecified())
                return;
        }
        sendPacket();
        scheduleNextPacket(false);
    }
    else
        llcSocket.processMessage(msg);
}

void EtherAppGoose::handleStartOperation(LifecycleOperation *operation)
{
    if (isGenerator())
        scheduleNextPacket(true);
}

void EtherAppGoose::handleStopOperation(LifecycleOperation *operation)
{
    cancelNextPacket();
    llcSocket.close();
    delayActiveOperationFinish(par("stopOperationTimeout"));
}

void EtherAppGoose::handleCrashOperation(LifecycleOperation *operation)
{
    cancelNextPacket();
    if (operation->getRootModule() != getContainingNode(this))
        llcSocket.destroy();
}

void EtherAppGoose::socketClosed(inet::Ieee8022LlcSocket *socket)
{
    if (operationalState == State::STOPPING_OPERATION && !llcSocket.isOpen())
        startActiveOperationExtraTimeOrFinish(par("stopOperationExtraTime"));
}

bool EtherAppGoose::isGenerator()
{
    return par("destAddress").stringValue()[0];
}

void EtherAppGoose::scheduleNextPacket(bool start)
{
    simtime_t cur = simTime();
    simtime_t next;
    if (start) {
        next = cur <= startTime ? startTime : cur;
        timerMsg->setKind(START);
    }
    else {
        next = cur + *sendInterval;
        timerMsg->setKind(NEXT);
    }
    if (stopTime < SIMTIME_ZERO || next < stopTime)
        scheduleAt(next, timerMsg);
}

void EtherAppGoose::cancelNextPacket()
{
    if (timerMsg)
        cancelEvent(timerMsg);
}

MacAddress EtherAppGoose::resolveDestMacAddress()
{
    MacAddress destMacAddress;
    const char *destAddress = par("destAddress");
    if (destAddress[0]) {
        if (!destMacAddress.tryParse(destAddress))
            destMacAddress = L3AddressResolver().resolve(destAddress, L3AddressResolver::ADDR_MAC).toMac();
    }
    return destMacAddress;
}

void EtherAppGoose::sendPacket()
{
    seqNum++;

    char msgname[30];
    sprintf(msgname, "req-%d-%ld", getId(), seqNum);
    EV_INFO << "Generating GOOSE packet `" << msgname << "'\n";

    Packet *datapacket = new Packet(msgname, IEEE802CTRL_DATA);
    const auto& data = makeShared<EtherAppReq>();
    data->addTag<CreationTimeTag>()->setCreationTime(simTime());
    //auto creationTimeTag = data->addTag<CreationTimeTag>();
    //creationTimeTag->setCreationTime(simTime());

    data->setRequestId(seqNum);

    long randomInterval = 0;
    randomInterval = (60-50)*( (double)rand() / (double)RAND_MAX);
    randomInterval = randomInterval + 50;

    long len = randomInterval;

    data->setChunkLength(B(len));

    long randomIntervalTwo = 0;
    randomIntervalTwo = (60-50)*( (double)rand() / (double)RAND_MAX);
    randomIntervalTwo = randomInterval + 50;

    long respLen = randomIntervalTwo;

    //long respLen = *respLength;
    data->setResponseBytes(respLen);
    datapacket->insertAtBack(data);

    datapacket->addTagIfAbsent<MacAddressReq>()->setDestAddress(destMacAddress);
    auto ieee802SapReq = datapacket->addTagIfAbsent<Ieee802SapReq>();
    ieee802SapReq->setSsap(localSap);
    ieee802SapReq->setDsap(remoteSap);

    EV_INFO << "Send GOOSE packet `" << msgname << "' dest=" << destMacAddress << " length=" << len << "\n";
    emit(packetSentSignal, datapacket);
    llcSocket.send(datapacket);
    packetsSent++;
}

void EtherAppGoose::socketDataArrived(Ieee8022LlcSocket*, Packet *msg)
{
    EV_INFO << "Received packet `" << msg->getName() << "'\n";

    packetsReceived++;

    auto data = msg->peekData();
    auto regions = data->getAllTags<CreationTimeTag>();

    for (auto& region : regions) { // for each region do
        auto creationTime = region.getTag()->getCreationTime(); // original time
        auto delay = simTime() - creationTime; // compute delay
        std::cout << region.getOffset() << region.getLength() << delay; // use data
    }

    emit(packetReceivedSignal, msg);

    delete msg;
}

void EtherAppGoose::finish()
{
    cancelAndDelete(timerMsg);
    timerMsg = nullptr;
}

} // namespace inet

